# Copyright 2017 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent.metadata import driver_base

_HEADER_CONFIG_TEMPLATE = """
    http-request add-header X-OVN-%(res_type)s-ID %(res_id)s
"""


class HaproxyConfigurator(driver_base.HaproxyConfiguratorBase):
    PROXY_CONFIG_DIR = "ovn-metadata-proxy"
    HEADER_CONFIG_TEMPLATE = _HEADER_CONFIG_TEMPLATE


class MetadataDriver(driver_base.MetadataDriverBase):
    @staticmethod
    def haproxy_configurator():
        return HaproxyConfigurator
